""" This is the execution manager. """

import logging
from typing import Dict

from messaging.router import Router, on_message

from workspaces.capability.enums import (
    CapabilityRequestState,
    CapabilityStepType,
    ExecutionPriority,
    ExecutionState,
)
from workspaces.capability.helpers import CapabilitySequence
from workspaces.capability.schema import CapabilityRequest
from workspaces.capability.schema_interfaces import CapabilityExecutionIF
from workspaces.capability.services.capability_queue import CapabilityQueue
from workspaces.capability.services.interfaces import CapabilityInfoIF
from workspaces.workflow.services.interfaces import WorkflowServiceIF

# pylint: disable=E0401, W1203


logger = logging.getLogger(__name__)


class ExecutionManager:
    """
    Class that manages the capability execution system. It's duties include:
    - Creating new executions
    - Managing existing execution state
    - Handling capability events and updating the state of the system accordingly
    """

    def __init__(
        self,
        capability_info: CapabilityInfoIF,
        workflow_service: WorkflowServiceIF,
        message_router: Router,
    ):
        self.execution_pool = []
        self.queues = {}
        self.capability_info = capability_info
        self.workflow_service = workflow_service
        self.message_router = message_router
        self.message_router.register(self)

    def run_capability(self, request: CapabilityRequest) -> CapabilityExecutionIF:
        """
        Run a capability with the requested settings
        TODO: Executions should be made from capability versions, not requests

        :param request: Capability request
        :return: Execution of the request
        """
        request.state = CapabilityRequestState.Submitted.name
        self.capability_info.save_entity(request)
        execution = self.capability_info.create_execution(request)
        steps = CapabilitySequence.from_str(execution.steps)

        if steps[0].step_type == CapabilityStepType.PrepareAndRunWorkflow:
            # First step is to run workflow, move to capability queue
            # FIXME: Priority needs to be dynamic
            self.enqueue_execution(execution, ExecutionPriority.Default.value)
        else:
            self.execution_pool.append(execution)
            execution.state = ExecutionState.Waiting.name
        return execution

    @on_message(service="workflow", type="workflow-executing")
    @on_message(service="workflow", type="workflow-complete")
    def send_step_complete(self, **message: Dict):
        """
        Send the step-complete message

        :param message: message to send
        :return:
        """
        logger.info(f"RECEIVED WORKFLOW MESSAGE: {message}")
        subject = message["subject"]
        execution = self.capability_info.lookup_execution_by_workflow_request_id(
            subject["workflow_request_id"]
        )
        step_complete_msg = {
            "service": "capability",
            "routing_key": "capability",
            "subject": execution.__json__(),
            "type": "step-complete",
        }
        self.message_router.send_message(**step_complete_msg)

    @on_message(service="workflow", type="workflow-failed")
    def on_workflow_failed(self, **message: Dict):
        """
        Announce and record workflow failure

        :param message: the failure message
        :return:
        """
        subject = message["subject"]
        execution = self.capability_info.lookup_execution_by_workflow_request_id(
            subject["workflow_request_id"]
        )

        execution.state = ExecutionState.Failed.name

        execution_failed_msg = {
            "service": "capability",
            "routing_key": "capability",
            "subject": execution.__json__(),
            "type": "execution-failed",
        }
        self.message_router.send_message(**execution_failed_msg)
        self.capability_info.save_execution(execution)

    @on_message(message="Ingestion complete")
    def on_ingestion_complete(self, **message: Dict):
        """
        Announce and record a completed ingestion event.

        :param message: the ingestion-complete message
        :return:
        """

        logger.info(f"This is a ingestion-complete message: {message}")

    @on_message(service="capability", type="execution-failed")
    def on_execution_fail(self, **message: Dict):
        """
        Announce and record execution failure

        :param message: the failure message
        :return:
        """
        subject = message["subject"]
        execution = self.capability_info.lookup_execution_by_workflow_request_id(
            subject["current_workflow_request_id"]
        )
        # Release capability engines on execution failure
        self.queues[execution.capability.name].complete_execution(execution.id)

    @on_message(service="capability", type="step-complete")
    def on_step_complete(self, **message: Dict):
        """
        Gets called when a capability execution completes a step; advances that capability to its
        next step

        :param message: AMQP step-complete message
        """
        logger.info(f"RECEIVED CAPABILITY MESSAGE: {message}")
        subject = message["subject"]
        if subject["type"] == "CapabilityExecution":
            execution_id = subject["id"]
            execution = self.capability_info.lookup_execution(execution_id)
        else:
            logger.error(f"Unrecognized message type: {subject['type']}.")
            return
        step_sequence = CapabilitySequence.from_str(execution.steps)
        current_step = step_sequence[execution.current_step]

        if current_step.step_type == CapabilityStepType.AwaitWorkflow:
            # Workflow completed; return capability engine to available state
            self.queues[execution.capability.name].complete_execution(execution_id)
            logger.info(">>> WORKFLOW COMPLETE! SETTING EXECUTION TO WAITING")
            execution.state = ExecutionState.Waiting.name

        if execution.on_last_step():
            # Execution is now complete
            execution = self._complete_execution(execution)
        else:
            execution = self._advance_execution(execution, step_sequence)
        self.capability_info.save_execution(execution)

    def _advance_execution(
        self, execution: CapabilityExecutionIF, step_sequence: CapabilitySequence
    ) -> CapabilityExecutionIF:
        """
        Perform necessary steps to advance an execution to its next step

        :param execution: Execution to advance
        :return: Execution with updated metadata
        """
        execution.current_step += 1

        current_step = step_sequence[execution.current_step]
        if current_step.step_type == CapabilityStepType.PrepareAndRunWorkflow:
            # Enqueue execution that is on a run workflow step
            # FIXME: Priority needs to be dynamic; perhaps a priority column in execution schema?
            self.enqueue_execution(execution, ExecutionPriority.Default.value)

        return execution

    def _complete_execution(self, execution: CapabilityExecutionIF) -> CapabilityExecutionIF:
        """
        Perform necessary steps to declare a capability execution complete

        :param execution: Execution to complete
        :return: Execution with updated metadata
        """
        execution.state = ExecutionState.Complete.name

        execution_complete_msg = {
            "service": "capability",
            "routing_key": "capability",
            "subject": execution.__json__(),
            "type": "execution-complete",
        }
        self.message_router.send_message(**execution_complete_msg)

        return execution

    def enqueue_execution(
        self,
        execution: CapabilityExecutionIF,
        priority: int = ExecutionPriority.Default.value,
    ):
        """
        Move execution record that is ready to execute a workflow into the appropriate capability
        queue

        :param execution: Capability execution record
        :param priority: Priority of the requested execution
        """
        # Get correct queue or initialize one
        queue = self.queues.get(
            execution.capability.name,
            CapabilityQueue(
                self.capability_info,
                self.workflow_service,
                execution.capability.max_jobs,
            ),
        )
        queue.enqueue(execution, priority)
        self.queues[execution.capability.name] = queue

        # Remove execution record from pool
        if execution in self.execution_pool:
            self.execution_pool.remove(execution)
