import logging
from typing import Dict

import requests
from aat_wrest.metadata_wrester import WrestWorkflowMetadata
from messaging.router import Router, on_message

from aat_wrest.utilities import MDDBConnector
from pycapo import CapoConfig
from requests import Response

from workspaces.products.services.interfaces import ArchiveServiceIF

logger = logging.getLogger(__name__)


class ArchiveService(ArchiveServiceIF):
    """
    The archive service: clients access this to interface with the archive services
    """

    def __init__(self):
        self.message_router = Router("archive")
        self.message_router.register(self)

    @on_message(message="Ingestion complete")
    def on_ingestion_complete(self, **message: Dict):
        """
        Accept a ingestion-complete message from the archive

        :param message: the ingestion-complete message
        """

        logdata = message["logData"]
        execblock_id = logdata["execblock_id"]
        fileset_id = logdata["fileset_id"]
        project_code = logdata["project_code"]

        wrester = WrestWorkflowMetadata(MDDBConnector(), spl=None, fileset_id=fileset_id)

        response = self._create_std_calibration_request(
            wrester.spl, execblock_id, fileset_id, project_code
        )

        logger.debug(
            "Got result %s with type %s and body %s",
            response,
            type(response),
            response.content,
        )
        return wrester.spl, execblock_id, fileset_id, project_code

    @staticmethod
    def _create_std_calibration_request(
        spl: str, execblock_id: str, fileset_id: str, project_code: str
    ) -> Response:
        """
        Creates a new standard calibration with information given from an ingestion-completed event

        :param wrester: WrestWorkflowMetadata used to get product locator from the archive
        :param execblock_id: Execution block ID
        :param fileset_id: File Set ID
        :param project_code: Project Code
        :return: HTTP Response
        """
        capability_url = (
            CapoConfig().settings("edu.nrao.archive.workspaces.CapabilitySettings").serviceUrl
        )
        analyst_email = (
            CapoConfig().settings("edu.nrao.archive.workspaces.NotificationSettings").analystEmail
        )
        parameters = {
            "parameters": {
                "product_locator": spl,
                "user_email": analyst_email,
                "execblock_id": execblock_id,
                "fileset_id": fileset_id,
                "project_code": project_code,
            }
        }

        return requests.post(
            f"{capability_url}/capability/std_calibration/request/create", json=parameters
        )
